﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Drawing;
using System.Windows.Forms;
using MikuMikuPlugin;
using DxMath;

namespace CamBone
{
    public class Class1:IResidentPlugin,ICanSavePlugin,IHaveUserControl
    {
        private MyUserControl myUserControl;
        private Hashtable loadSaveParam = new Hashtable();

        public Guid GUID
        {
            get { return new Guid("B6819352-3ED5-4AD5-9A11-72C8B7A0388C"); }
        }
        public IWin32Window ApplicationForm { get; set; }
        public Scene Scene { get; set; }
        public string Description
        {
            get { return "CamBone Plugin v0.03 by Usagi"; }
        }
        public Image Image
        {
            get { return Properties.Resources.image; }
        }
        public Image SmallImage
        {
            get { return Properties.Resources.smallimage; }
        }
        public string Text
        {
            get { return "ボーンからの" + System.Environment.NewLine + "カメラ"; }
        }
        public string EnglishText
        {
            get { return "from Bone" + System.Environment.NewLine + "Camera"; }
        }

        public UserControl CreateControl()
        {
            myUserControl = new MyUserControl();
            return myUserControl;
        }

        // セーブとロード //
        private void parseParam(Stream stream)
        {
            using (StreamReader r = new StreamReader(stream))
            {
                string ln;
                char[] sep = new char[1];
                sep[0] = ':';
                while ((ln = r.ReadLine()) != null)
                {
                    string[] ss = ln.Split(sep, 2);
                    if (ss.Length >= 2)
                    {
                        loadSaveParam[ss[0]] = ss[1];
                    }
                }
            }

            myUserControl.TargetModelName = loadSaveParam["TargetModelName"].ToString();
            myUserControl.TargetBoneName = loadSaveParam["TargetBoneName"].ToString();
        }
        private string makeParam()
        {
            StringBuilder sb = new StringBuilder();
            Hashtable ht = new Hashtable();

            ht["TargetModelName"] = myUserControl.TargetModelName;
            ht["TargetBoneName"] = myUserControl.TargetBoneName;

            foreach (DictionaryEntry de in ht)
            {
                sb.AppendLine(string.Format("{0}:{1}", de.Key, de.Value));
            }
            return sb.ToString();
        }

        public Stream OnSaveProject()
        {
            Stream ret = new MemoryStream();
            string param = makeParam();
            byte[] buf = Encoding.UTF8.GetBytes(param);
            ret.Write(buf, 0, buf.Length);
            return ret;
        }

        public void OnLoadProject(Stream stream)
        {
            parseParam(stream);
        }

        public void Initialize()
        {
            return;
        }

        public void Enabled()
        {
        }

        public void Update(float Frame,
            float ElapsedTime)
        {
// カメラ計算 //
            // モデル探索
            Model targetModel = null;
            foreach (Model m in Scene.Models)
            {
                if (m.Name == myUserControl.TargetModelName)
                {
                    targetModel = m;
                    break;
                }
            }

            Model model = targetModel;
            Bone targetBone = null;
            MotionLayer targetLayer = null;
            bool flag = true;
            if (model != null)
            {
                targetBone = model.Bones[myUserControl.TargetBoneName];
                if (targetBone != null)
                {
                    foreach (MotionLayer ml in targetBone.Layers)
                    {
                        targetLayer = ml;
                        break;
                    }
                }
            }

            myUserControl.ModelFound = (targetModel != null) ? "有" : "無";
            myUserControl.BoneFound = (targetBone != null) ? "有" : "無";

            if (targetLayer == null)
            {
                flag = false;
            }

            if (flag)
            {
                MotionData cambone = targetLayer.CurrentLocalMotion;
                Vector3 c = cambone.Move;
                Quaternion rot = cambone.Rotation;
                Quaternion rotc = rot;
                rotc.Conjugate();
                Quaternion qr = new Quaternion(Vector3.UnitX, 0);
                Quaternion qt = new Quaternion(Vector3.UnitY, 0);
                Quaternion qf = new Quaternion(Vector3.UnitZ, 0);

                qr = rot * qr * rotc;
                qt = rot * qt * rotc;
                qf = rot * qf * rotc;

                float rate = 50.0f;
                Vector3 at = c + new Vector3(qf.X * rate,qf.Y * rate,qf.Z * rate);

                // 逆を向ける
                Quaternion rightV = new Quaternion(-qr.Axis,0);
                Quaternion upV = qt; // トップベクトルはそのまま
                Quaternion foreV = new Quaternion(-qf.Axis,0);

                Vector3 ang = calcMMMCam(rightV.Axis, upV.Axis, foreV.Axis);
                ang.X = -ang.X;
                ang.Y = -ang.Y;

                foreach (Camera cam in Scene.Cameras)
                {
                    foreach (CameraLayer ml in cam.Layers)
                    {
                        CameraMotionData cmd = ml.CurrentLocalMotion;
                        cmd.Position = at;
                        cmd.Angle = ang;
                        cmd.Radius = rate;
                        ml.CurrentLocalMotion = cmd;

                        break; // 先頭レイヤにのみ反映する
                    }

                    bool isMoving = false;
                    switch (Scene.State)
                    {
                        case SceneState.Playing:
                        case SceneState.PhysicsBaking:
                            isMoving = true;
                            break;
                    }
                    if (isMoving)
                    {
                        CameraMotionData cmd = cam.CurrentMotion;
                        cmd.Position = at;
                        cmd.Angle = ang;
                        cmd.Radius = rate;
                        cam.CurrentMotion = cmd;
                    }

                    break; // 先頭カメラにのみ反映する
                }

            }
        }

        public void Disabled()
        {
            myUserControl.ModelFound = "-";
            myUserControl.BoneFound = "-";
        }

        public void Dispose()
        {
        }

        /**
         * 姿勢を表現するベクトルから
         * カメラの姿勢を変更する角度を取得する。
         */
        public Vector3 calcMMMCam(Vector3 inRight,
            Vector3 inUp,
            Vector3 inFore)
        {
            Quaternion r = new Quaternion(inRight, 0);
            Quaternion t = new Quaternion(inUp, 0);
            Quaternion f = new Quaternion(inFore, 0);

            // 方角
            double dr = Math.Sqrt(f.X * f.X + f.Z * f.Z);
            double direction = 0;

            if (dr >= 0.001)
            {
                direction = Math.Atan2(f.X, f.Z);
            }

            Quaternion rot = Quaternion.RotationAxis(Vector3.UnitY, (float)-direction);
            Quaternion rotc = rot;
            rotc.Conjugate();

            r = rot * r * rotc;
            t = rot * t * rotc;
            f = rot * f * rotc;

            // 仰角
            double elevation = -Math.Atan2(f.Y, f.Z);

            rot = Quaternion.RotationAxis(Vector3.UnitX, (float)-elevation);
            rotc = rot;
            rotc.Conjugate();

            r = rot * r * rotc;
            t = rot * t * rotc;
            f = rot * f * rotc;

            // ねじり
            double ang3 = -Math.Atan2(t.X, t.Y);

            rot = Quaternion.RotationAxis(Vector3.UnitZ, (float)-ang3);
            rotc = rot;
            rotc.Conjugate();

            r = rot * r * rotc;
            t = rot * t * rotc;
            f = rot * f * rotc;

// ここまで単位行列になる
            /*int checks = 0;
            if (r.Axis.X > 0.999)
            {
                checks += 100;
            }
            if (t.Axis.Y > 0.999)
            {
                checks += 10;
            }
            if (f.Axis.Z > 0.999)
            {
                checks += 1;
            }
            myUserControl.TB4 = "" + checks;*/

            return new Vector3((float)elevation, (float)direction, (float)ang3);
        }

    }
    
}
